"""@package docstring
Iso2Mesh for Python - File I/O module

Copyright (c) 2024-2025 Qianqian Fang <q.fang at neu.edu>
"""

__all__ = [
    "affinemap",
    "meshinterp",
    "meshremap",
    "proj2mesh",
    "dist2surf",
    "regpt2surf",
]

##====================================================================================
## dependent libraries
##====================================================================================

import numpy as np
from iso2mesh.trait import nodesurfnorm
from iso2mesh.line import linextriangle

##====================================================================================
## implementations
##====================================================================================


def affinemap(pfrom, pto):
    """
    Calculate an affine transform (A matrix and b vector) to map 3D vertices
    from one space to another using least square solutions.

    Parameters:
    pfrom : numpy array (n x 3), points in the original space
    pto : numpy array (n x 3), points in the mapped space

    Returns:
    A : numpy array (3 x 3), the affine transformation matrix
    b : numpy array (3 x 1), the translation vector
    """
    ptnum = pfrom.shape[0]

    if pto.shape[0] != ptnum:
        raise ValueError("Two inputs should have the same size")

    bsubmat = np.eye(3)
    amat = np.zeros((ptnum * 3, 9))

    for i in range(ptnum):
        amat[i * 3 : (i + 1) * 3, :] = np.kron(bsubmat, pfrom[i, :])

    amat = np.hstack([amat, np.tile(bsubmat, (ptnum, 1))])
    bvec = pto.T.flatten()

    x, _, _, _ = np.linalg.lstsq(amat, bvec, rcond=None)
    A = x[:9].reshape(3, 3).T
    b = x[-3:]

    return A, b


def meshinterp(fromval, elemid, elembary, fromelem, initval=None):
    """
    Interpolate nodal values from the source mesh to the target mesh based on a linear interpolation.

    Args:
        fromval: Values defined at the source mesh nodes. The row or column number
                 must be the same as the source mesh node count (matching elemid length).
        elemid: IDs of the source mesh element that encloses the target mesh nodes; a vector of length
                equal to the target mesh node count.
        elembary: Barycentric coordinates of each target mesh node within the source mesh elements.
                  The sum of each row is 1, with 3 or 4 columns.
        fromelem: The element list of the source mesh.
        initval: Optional initial values for the target nodes.

    Returns:
        newval: A 2D array where rows equal the target mesh nodes, and columns equal
                the value numbers defined at each source mesh node.
    """

    # If fromval is a single row, convert it to a column vector
    if fromval.shape[0] == 1:
        fromval = fromval.T

    # Find valid indices (non-NaN element IDs)
    idx = np.where(~np.isnan(elemid))[0]

    # Reshape fromval to match the size of elembary and fromelem
    allval = np.reshape(
        fromval[fromelem[elemid[idx], :], :],
        (len(idx), elembary.shape[1], fromval.shape[1]),
    )

    # Perform linear interpolation using barycentric coordinates
    tmp = np.array(
        [np.sum(elembary[idx, :] * x, axis=1) for x in np.rollaxis(allval, 2)]
    )

    # Initialize newval with initval or NaN
    if initval is not None:
        newval = initval
    else:
        newval = np.full((len(elemid), fromval.shape[1]), np.nan)

    # Assign interpolated values to newval at valid indices
    newval[idx, :] = np.squeeze(np.stack(tmp, axis=2))

    return newval


def meshremap(fromval, elemid, elembary, toelem, nodeto):
    """
    Redistribute nodal values from the source mesh to the target mesh
    so that the sum of each property on each mesh is the same.

    Parameters:
    fromval: Values defined at the source mesh nodes; should be a 1D or 2D array
             with the same number of rows or columns as the source mesh node count.
    elemid: IDs of the target mesh element that encloses the source mesh nodes; a vector.
    elembary: Barycentric coordinates of each source mesh node within the target mesh elements;
              sum of each row is 1, expect 3 or 4 columns (or N-D).
    toelem: Element list of the target mesh.
    nodeto: Total number of target mesh nodes.

    Returns:
    newval: A 2D array with rows equal to the target mesh nodes and columns equal to
            the value numbers defined at each source mesh node.
    """

    # Ensure fromval is a column vector if it is a row vector
    if fromval.shape[0] == 1:
        fromval = fromval.T

    # Ensure fromval's number of columns matches the length of elemid
    if fromval.shape[1] == len(elemid):
        fromval = fromval.T

    # Initialize output array newval
    newval = np.zeros((nodeto, fromval.shape[1]))

    # Ignore NaN elements
    idx = ~np.isnan(elemid)
    fromval = fromval[idx, :]
    elembary = elembary[idx, :]
    idx = elemid[idx]

    # Compute nodal values weighted by barycentric coordinates
    nodeval = np.repeat(
        fromval[:, np.newaxis, :], elembary.shape[1], axis=1
    ) * np.repeat(elembary[:, np.newaxis, :], fromval.shape[1], axis=1).transpose(
        0, 2, 1
    )

    # Accumulate contributions to target mesh nodes
    for i in range(elembary.shape[1]):
        ix, iy = np.meshgrid(toelem[idx, i], np.arange(fromval.shape[1]))
        nval = nodeval[:, :, i].T
        newval += np.add.at(newval, (ix.flatten(), iy.flatten()), nval.flatten())

    return newval


def proj2mesh(v, f, pt, nv=None, cn=None, radmax=None):
    """
    Project a point cloud onto the surface mesh (triangular surface only).

    Parameters:
        v : ndarray
            Node coordinates of the surface mesh (nn x 3)
        f : ndarray
            Element list of the surface mesh (triangular or cubic, use only 3 columns for triangle)
        pt : ndarray
            Points to be projected, with 3 columns for x, y, and z respectively
        nv : ndarray, optional
            Nodal normals (size: v.shape[0] x 3), calculated by nodesurfnorm
        cn : ndarray, optional
            Integer vector of closest surface node indices for each point in pt (from dist2surf)
        radmax : float, optional
            If specified, limits search for elements to those within a bounding box centered at the point

    Returns:
        newpt : ndarray
            Projected points from pt
        elemid : ndarray
            Indices of the surface triangle that contains each projected point
        weight : ndarray
            Barycentric coordinates (weights) for each projected point

    -- this function is part of iso2mesh toolbox (http://iso2mesh.sf.net)
    """
    cent = np.mean(v, axis=0)
    enum = len(f)
    ec = np.reshape(v[f[:, :3] - 1].transpose(1, 2, 0), (3, 3, enum))
    centroid = np.mean(ec, axis=1)
    newpt = np.zeros_like(pt)
    elemid = np.zeros(pt.shape[0], dtype=int)
    weight = np.zeros((pt.shape[0], 3))

    idoldmesh = np.any(np.all(pt[:, None, :] == v[None, :, :], axis=2), axis=1)
    idnode = np.where(idoldmesh)[0]
    if idnode.size > 0:
        for idx in idnode:
            matches = np.where(
                np.all(
                    f[:, None, :] == np.where((v == pt[idx]).all(axis=1))[0][0] + 1,
                    axis=2,
                )
            )
            if matches[0].size > 0:
                newpt[idx, :] = pt[idx, :]
                elemid[idx] = matches[0][0] + 1
                weight[idx, matches[1][0]] = 1

    if nv is not None and cn is not None:
        direction = nv[cn - 1, :]
        if radmax is not None:
            radlimit = radmax
        else:
            radlimit = -1
    else:
        direction = pt - cent
        radlimit = -1

    for t in range(pt.shape[0]):
        if idoldmesh[t]:
            continue

        maxdist = np.linalg.norm(pt[t, :] - cent)
        if radlimit > 0:
            maxdist = radlimit

        mask = np.all(np.abs(centroid.T - pt[t]) < maxdist, axis=1)
        idx = np.where(mask)[0]
        dist = centroid[:, idx] - pt[t, :, None]
        c0 = np.sum(dist**2, axis=0)

        sorted_idx = np.argsort(c0)

        for i in range(len(idx)):
            inside, p, w = linextriangle(
                pt[t, :],
                pt[t, :] + direction[t, :],
                v[f[idx[sorted_idx[i]], :3] - 1, :],
            )
            if inside:
                newpt[t, :] = p
                weight[t, :] = w
                elemid[t] = idx[sorted_idx[i]] + 1
                break

    return newpt, elemid, weight


def dist2surf(node, nv, p, cn=None):
    """
    Calculate the distances from a point cloud to a surface, and return
    the indices of the closest surface node.

    Parameters:
        node : ndarray
            Node coordinates of the surface mesh (nn x 3).
        nv : ndarray
            Nodal normals (vector) calculated from nodesurfnorm(), shape (nn x 3).
        p : ndarray
            Points to be calculated, shape (N x 3).
        cn : ndarray, optional
            If provided, an integer vector of indices of the closest surface nodes.

    Returns:
        d2surf : ndarray
            Distances from each point in p to the surface.
        cn : ndarray
            Indices of the closest surface nodes.

    -- this function is part of "metch" toolbox, see COPYING for license
    """

    if cn is None:
        nn = node.shape[0]
        pnum = p.shape[0]
        mindist = np.zeros(pnum)
        cn = np.zeros(pnum, dtype=int)
        for i in range(pnum):
            d0 = node - np.tile(p[i, :], (nn, 1))
            d0 = np.sum(d0 * d0, axis=1)
            cn[i] = np.argmin(d0)
            mindist[i] = d0[cn[i]]
    d2surf = np.abs(np.sum(nv[cn, :] * (p - node[cn, :]), axis=1))

    return d2surf, cn


def regpt2surf(node, elem, p, pmask, A0, b0, cmask, maxiter):
    """
    Perform point cloud registration to a triangular surface
    (surface can be either triangular or cubic), using Gauss-Newton method.

    Parameters:
        node : ndarray
            Node coordinates of the surface mesh (nn x 3).
        elem : ndarray
            Element list of the surface mesh (triangular: 3 columns, cubic: 4 columns).
        p : ndarray
            Points to be registered (N x 3).
        pmask : ndarray
            Mask vector of same length as p. If pmask[i] == -1, the point is free;
            if 0, it is fixed; if n > 0, its distance to node[n-1] is minimized.
        A0 : ndarray
            Initial guess for affine A matrix (3x3).
        b0 : ndarray
            Initial guess for affine b vector (3,).
        cmask : ndarray
            Binary vector of length 12, indicates which of [A.flatten(); b] to optimize.
        maxiter : int
            Maximum number of optimization iterations.

    Returns:
        A : ndarray
            Updated affine transformation matrix (3x3).
        b : ndarray
            Updated translation vector (3,).
        newpos : ndarray
            Transformed positions of input points.
    """

    A = A0.copy()
    b = b0.reshape(-1)

    # Wrap A and b into single vector C
    C = np.concatenate([A.flatten(), b])
    delta = 1e-4

    newpos = (C[:9].reshape(3, 3) @ p.T + C[9:].reshape(3, 1)).T
    nv = nodesurfnorm(node, elem)

    clen = len(C)
    cuplist = np.where(cmask == 1)[0]
    pfree = np.where(pmask < 0)[0]
    pfix = np.where(pmask > 0)[0]

    for iter in range(maxiter):
        dist0 = np.zeros(len(pfree) + len(pfix))

        if len(pfree) > 0:
            dist0[pfree], cn0 = dist2surf(node, nv, newpos[pfree])
        else:
            cn0 = []

        if len(pfix) > 0:
            fixdist = node[pmask[pfix] - 1] - newpos[pfix]
            dist0[pfix] = np.sqrt(np.sum(fixdist**2, axis=1))

        print(f"iter={iter+1} error={np.sum(np.abs(dist0))}")

        J = np.zeros((len(dist0), clen))
        for i in range(clen):
            if cmask[i] == 0:
                continue
            dC = C.copy()
            dC[i] = C[i] * (1 + delta) if C[i] != 0 else C[i] + delta
            newp = (dC[:9].reshape(3, 3) @ p.T + dC[9:].reshape(3, 1)).T

            dist = np.zeros(len(dist0))
            if len(pfree) > 0:
                if len(cn0) == len(pfree):
                    dist[pfree], _ = dist2surf(node, nv, newp[pfree], cn0)
                else:
                    dist[pfree], _ = dist2surf(node, nv, newp[pfree])
            if len(pfix) > 0:
                fixdist = node[pmask[pfix] - 1] - newp[pfix]
                dist[pfix] = np.sqrt(np.sum(fixdist**2, axis=1))

            J[:, i] = (dist - dist0) / (dC[i] - C[i])

        wj = np.sqrt(np.sum(J**2, axis=0))
        J[:, cuplist] = J[:, cuplist] / wj[cuplist]

        dC = np.linalg.lstsq(J[:, cuplist], dist0, rcond=None)[0] / wj[cuplist]
        C[cuplist] -= 0.5 * dC

        newpos = (C[:9].reshape(3, 3) @ p.T + C[9:].reshape(3, 1)).T

    A = C[:9].reshape(3, 3)
    b = C[9:]
    return A, b, newpos



"""@package docstring
Iso2Mesh for Python - Polyline processing utilities

Copyright (c) 2024-2025 Qianqian Fang <q.fang at neu.edu>
"""

__all__ = [
    "linextriangle",
    "getplanefrom3pt",
    "polylinelen",
    "closestnode",
    "polylineinterp",
    "polylinesimplify",
]
##====================================================================================
## dependent libraries
##====================================================================================

import numpy as np

##====================================================================================
## implementations
##====================================================================================


def linextriangle(p0, p1, plane):
    """
    Calculate the intersection of a 3D line (defined by two points) with a plane (defined by 3 points).

    Parameters:
        p0 : array_like
            A 3D point in the form of (x, y, z)
        p1 : array_like
            Another 3D point, p0 and p1 determine the line
        plane : ndarray
            A 3x3 matrix, each row is a 3D point defining the triangle on the plane

    Returns:
        isinside : bool
            True if the intersection is within the triangle defined by the plane; False otherwise
        pt : ndarray
            Coordinates of the intersection point
        coord : ndarray
            Barycentric coordinates of the intersection point (if inside); otherwise all zeros.

    -- this function is part of iso2mesh toolbox (http://iso2mesh.sf.net)
    """

    a, b, c, d = getplanefrom3pt(plane)

    if a**2 + b**2 + c**2 == 0.0:
        raise ValueError("Degenerated plane")

    dl_n = np.dot([a, b, c], np.array(p1) - np.array(p0))

    if dl_n == 0.0:
        raise ValueError("Degenerated line")

    # Solve for the intersection point
    t = -(a * p0[0] + b * p0[1] + c * p0[2] + d) / dl_n
    pt = np.array(p0) + t * (np.array(p1) - np.array(p0))

    dist = np.sum(np.abs(np.diff(plane, axis=0)), axis=0)
    imax = np.argsort(dist)
    if dist[imax[1]] == 0.0:
        raise ValueError("Degenerated triangle")

    goodidx = imax[1:]

    ptproj = pt[goodidx]
    mat0 = np.vstack([plane[:, goodidx].T, ptproj, np.ones(4)])

    isinside = False
    coord = np.array([0.0, 0.0, 0.0])

    det1 = np.linalg.det(mat0[:, [3, 1, 2]])
    det2 = np.linalg.det(mat0[:, [0, 3, 2]])
    if det1 * det2 < 0:
        return isinside, pt, coord

    det3 = np.linalg.det(mat0[:, [0, 1, 3]])
    if det2 * det3 < 0:
        return isinside, pt, coord
    if det1 * det3 < 0:
        return isinside, pt, coord

    isinside = True
    det0 = np.linalg.det(mat0[:, [0, 1, 2]])
    coord = np.array([det1, det2, det3]) / det0

    return isinside, pt, coord


def getplanefrom3pt(plane):
    """
    Define a plane equation ax + by + cz + d = 0 from three 3D points.

    Parameters:
        plane : ndarray
            A 3x3 matrix with each row specifying a 3D point (x, y, z).

    Returns:
        a, b, c, d : float
            The coefficients of the plane equation ax + by + cz + d = 0.

    -- this function is part of iso2mesh toolbox (http://iso2mesh.sf.net)
    """

    x = plane[:, 0]
    y = plane[:, 1]
    z = plane[:, 2]

    # Compute the plane equation a*x + b*y + c*z + d = 0
    a = y[0] * (z[1] - z[2]) + y[1] * (z[2] - z[0]) + y[2] * (z[0] - z[1])
    b = z[0] * (x[1] - x[2]) + z[1] * (x[2] - x[0]) + z[2] * (x[0] - x[1])
    c = x[0] * (y[1] - y[2]) + x[1] * (y[2] - y[0]) + x[2] * (y[0] - y[1])
    d = -np.linalg.det(plane)

    return a, b, c, d


def closestnode(node, p):
    """
    Find the closest point in a node list and return its index.

    Parameters:
        node : ndarray
            Each row is an N-D node coordinate.
        p : ndarray
            A given position in the same space (1D array).

    Returns:
        idx : int
            The index of the position in the node list that has the shortest
            Euclidean distance to the position p.
        dist : float
            The squared distance between p and the closest node.

    -- this function is part of brain2mesh toolbox (http://mcx.space/brain2mesh)
       License: GPL v3 or later, see LICENSE.txt for details
    """

    dd = node - np.tile(p, (node.shape[0], 1))
    dist_sq = np.sum(dd * dd, axis=1)
    idx = np.argmin(dist_sq)
    dist = dist_sq[idx]
    return idx + 1, dist


def polylinelen(node, p0=None, p1=None, pmid=None):
    """
    Calculate the polyline line segment length vector in sequential order

    Parameters:
        node : ndarray
            An N x 3 array defining each vertex of the polyline in sequential order.
        p0 : int or ndarray, optional
            A given node to define the start of the polyline (index or coordinate).
            If not defined, start position is assumed to be the first node.
        p1 : int or ndarray, optional
            A given node to define the end of the polyline (index or coordinate).
            If not defined, end position is assumed to be the last node.
        pmid : int or ndarray, optional
            A given node that sits between p0 and p1. If not defined, the floor of the middle index is used.

    Returns:
        len : ndarray
            The length of each segment between the start and end points.
        node : ndarray
            The node list between the start and end points of the polyline.
        inputreversed : bool
            1 if the input node is reversed from p0 to pmid to p1, 0 otherwise.

    -- this function is part of brain2mesh toolbox (http://mcx.space/brain2mesh)
       License: GPL v3 or later, see LICENSE.txt for details
    """

    if p1 is None:
        p1 = node.shape[0]
        if p0 is None:
            p0 = 1

    if pmid is None:
        pmid = np.floor((p0 + p1) * 0.5)

    if isinstance(p0, (list, np.ndarray)) and np.shape(p0)[-1] == 3:
        p0, _ = closestnode(node, np.asarray(p0))
    if isinstance(p1, (list, np.ndarray)) and np.shape(p1)[-1] == 3:
        p1, _ = closestnode(node, np.asarray(p1))
    if isinstance(pmid, (list, np.ndarray)) and np.shape(pmid)[-1] == 3:
        pmid, _ = closestnode(node, np.asarray(pmid))

    # convert to 0-based indices
    p0 -= 1
    p1 -= 1
    pmid -= 1

    if p0 < pmid and pmid < p1:
        inputreversed = 0
        node = node[range(p0, p1 + 1), :]
    elif p0 < pmid and p1 < pmid:
        inputreversed = min(p0, p1) == p0
        idx_range = list(range(min(p0, p1), -1, -1)) + list(
            range(node.shape[0] - 1, max(p0, p1) - 1, -1)
        )
        node = node[idx_range, :]
        if not inputreversed:
            node = np.flipud(node)
    elif p0 > pmid and pmid > p1:
        inputreversed = 1
        node = node[range(p0, p1 - 1, -1), :]
    elif p0 > pmid and p1 > pmid:
        inputreversed = max(p0, p1) == p1
        idx_range = list(range(max(p0, p1), node.shape[0])) + list(
            range(0, min(p0, p1) + 1)
        )
        node = node[idx_range, :]
        if inputreversed:
            node = np.flipud(node)

    seg = node[:-1, :] - node[1:, :]
    length = np.sqrt(np.sum(seg * seg, axis=1))
    return length, node, inputreversed


def polylineinterp(polylen, length, nodes=None):
    """
    Find the polyline segment indices and interpolation weights for a
    specified total length or a set of lengths.

    Parameters:
        polylen : array_like
            A 1D vector sequentially recording the length of each segment
            of a polyline. The first number is the length of the 1st segment, and so on.
        length : float or array_like
            A scalar or array specifying the total length(s) to interpolate.
        nodes : ndarray, optional
            If provided, should be an array with shape (len(polylen)+1, N),
            where each row is a coordinate for a node along the polyline.

    Returns:
        idx : ndarray
            The indices (1-based) of the polyline segments where each length ends.
            NaN if length > sum(polylen). Indexing starts at 1 (like MATLAB).
        weight : ndarray
            Interpolation weights (0-1) toward the end node of the segment.
        newnodes : ndarray, optional
            Interpolated node positions corresponding to `length`.

    Example:
        polylen = [2, 2, 1, 7, 10]
        idx, weight = polylineinterp(polylen, [3, 12, 7])
    """

    polylen = np.asarray(polylen, dtype=float).flatten()
    length = np.atleast_1d(length).astype(float)
    cumlen = np.concatenate(([0], np.cumsum(polylen)))

    idx = np.full(length.shape, np.nan)
    weight = np.zeros(length.shape)

    compute_nodes = nodes is not None and nodes.shape[0] == len(polylen) + 1
    newnodes = np.zeros((len(length), nodes.shape[1])) if compute_nodes else None

    for i in range(len(length)):
        pos = np.histogram(length[i], bins=cumlen)[0]
        if np.any(pos == 1):
            seg = np.searchsorted(cumlen, length[i], side="right")
            if seg == len(cumlen):
                idx[i] = seg - 1
                weight[i] = 1.0
                if compute_nodes:
                    newnodes[i, :] = nodes[-1, :]
            elif seg <= len(polylen):
                idx[i] = seg
                weight[i] = (length[i] - cumlen[seg - 1]) / polylen[seg - 1]
                if compute_nodes:
                    newnodes[i, :] = (1 - weight[i]) * nodes[seg - 1, :] + weight[
                        i
                    ] * nodes[seg, :]

    idx[idx > len(polylen)] = np.nan

    return idx, weight, newnodes


def polylinesimplify(nodes, minangle=None):
    """
    Calculate a simplified polyline by removing nodes where two adjacent
    segments have an angle less than a specified limit.

    Parameters:
        nodes : ndarray
            An N x 3 array defining each vertex of the polyline in sequential order.
        minangle : float, optional
            Minimum segment angle in radians. If not given, defaults to 0.75 * pi.

    Returns:
        newnodes : ndarray
            The updated node list; start/end will not be removed.
        len : ndarray
            The length of each segment between the start and end points.

    -- this function is part of brain2mesh toolbox (http://mcx.space/brain2mesh)
    """

    if minangle is None:
        minangle = 0.75 * np.pi

    def segvec(n1, n2):
        v = n2 - n1
        normals = np.linalg.norm(v, axis=1, keepdims=True)
        return v / normals

    v = segvec(nodes[:-1], nodes[1:])
    dotprod = np.sum(-v[:-1] * v[1:], axis=1)
    ang = np.arccos(np.clip(dotprod, -1.0, 1.0))

    newnodes = nodes.copy()
    newv = v.copy()
    newang = ang.copy()

    idx = np.where(newang < minangle)[0]

    while len(idx) > 0:
        newnodes = np.delete(newnodes, idx + 1, axis=0)
        newv = np.delete(newv, idx + 1, axis=0)
        newang = np.delete(newang, idx, axis=0)

        idx = idx - np.arange(len(idx))
        idx = np.unique(idx)

        idx1 = idx[idx < newnodes.shape[0] - 1]
        if len(idx1) > 0:
            newv[idx1, :] = segvec(newnodes[idx1], newnodes[idx1 + 1])
        idx1 = idx[idx < newv.shape[0] - 1]
        if len(idx1) > 0:
            newang[idx1] = np.arccos(np.sum(-newv[idx1] * newv[idx1 + 1], axis=1))
        idx0 = idx[idx > 0]
        if len(idx0) > 0:
            newang[idx0 - 1] = np.arccos(np.sum(-newv[idx0 - 1] * newv[idx0], axis=1))

        idx = np.where(newang < minangle)[0]

    if newnodes.shape[0] > 1:
        lenvec = newnodes[:-1] - newnodes[1:]
        length = np.sqrt(np.sum(lenvec**2, axis=1))
    else:
        length = np.array([])

    return newnodes, length



"""@package docstring
Iso2Mesh for Python - Primitive shape meshing functions

Copyright (c) 2024 Edward Xu <xu.ed at neu.edu>
              2024-2025 Qianqian Fang <q.fang at neu.edu>
"""

__all__ = [
    "meshgrid5",
    "meshgrid6",
    "latticegrid",
    "meshabox",
    "meshacylinder",
    "meshcylinders",
    "meshcylinders",
    "meshanellip",
    "meshunitsphere",
    "meshasphere",
    "extrudecurve",
    "extrudesurf",
]

##====================================================================================
## dependent libraries
##====================================================================================

import numpy as np
import sys
from itertools import permutations, combinations
from iso2mesh.core import surf2mesh, vol2restrictedtri
from iso2mesh.trait import meshreorient, volface, surfedge, nodesurfnorm
from iso2mesh.utils import *
from iso2mesh.modify import removeisolatednode, meshcheckrepair

# _________________________________________________________________________________________________________


def meshabox(p0, p1, maxvol=None, nodesize=1, **kwargs):
    """
    Create the surface and tetrahedral mesh of a box geometry.

    Parameters:
    p0: Coordinates (x, y, z) for one end of the box diagonal
    p1: Coordinates (x, y, z) for the other end of the box diagonal
    opt: Maximum volume of the tetrahedral elements
    nodesize: (Optional) Size of the elements near each vertex.
              Can be a scalar or an 8x1 array.

    Returns:
    node: Node coordinates, 3 columns for x, y, and z respectively
    face: Surface mesh faces, each row represents a face element
    elem: Tetrahedral elements, each row represents a tetrahedron
    """
    if nodesize is None:
        nodesize = 1

    # Call to surf2mesh function to generate the surface mesh and volume elements
    node, elem, ff = surf2mesh(
        np.array([]),
        np.array([]),
        p0,
        p1,
        1,
        maxvol,
        regions=None,
        holes=None,
        dobbx=nodesize,
        **kwargs,
    )

    # Reorient the mesh elements
    elem, _, _ = meshreorient(node, elem[:, :4])

    # Extract the surface faces from the volume elements
    face = volface(elem)[0]

    return node, face, elem


# _________________________________________________________________________________________________________


def meshunitsphere(tsize, maxvol=None):
    dim = 60
    esize = tsize * dim
    thresh = dim / 2 - 1

    xi, yi, zi = np.meshgrid(
        np.arange(0, dim + 0.5, 0.5),
        np.arange(0, dim + 0.5, 0.5),
        np.arange(0, dim + 0.5, 0.5),
    )
    dist = thresh - np.sqrt((xi - 30) ** 2 + (yi - 30) ** 2 + (zi - 30) ** 2)
    dist[dist < 0] = 0

    # Call a vol2restrictedtri equivalent in Python here (needs a custom function)
    node, face = vol2restrictedtri(
        dist, 1, (dim, dim, dim), dim**3, 30, esize, esize, 40000
    )

    node = (node - 0.5) * 0.5
    node, face, _ = removeisolatednode(node, face)

    node = (node - 30) / 28
    r0 = np.sqrt(np.sum(node**2, axis=1))
    node = node / r0[:, None]

    if not maxvol:
        maxvol = tsize**3

    # Call a surf2mesh equivalent in Python here (needs a custom function)
    node, elem, face = surf2mesh(
        node, face, np.array([-1, -1, -1]) * 1.1, np.array([1, 1, 1]) * 1.1, 1, maxvol
    )

    return node, face, elem


# _________________________________________________________________________________________________________


def meshasphere(c0, r, tsize, maxvol=None):
    if maxvol is None:
        maxvol = tsize**3

    if maxvol is not None:
        node, face, elem = meshunitsphere(tsize / r, maxvol=maxvol / (r**3))
    else:
        node, face, elem = meshunitsphere(tsize / r)

    node = node * r + np.tile(np.array(c0).reshape(1, -1), (node.shape[0], 1))

    return node, face, elem  # if maxvol is not None else (node, face)


# _________________________________________________________________________________________________________


def meshacylinder(c0, c1, r, tsize=None, maxvol=None, ndiv=20):
    if len(np.array([r])) == 1:
        r = np.array([r, r])

    if np.any(np.array(r) <= 0) or np.all(c0 == c1):
        raise ValueError("Invalid cylinder parameters")

    c0 = np.array(c0).reshape(-1, 1)
    c1 = np.array(c1).reshape(-1, 1)
    v0 = c1 - c0
    len_axis = np.linalg.norm(v0)

    if tsize is None:
        tsize = min(np.append(r, len_axis)) / 10

    if maxvol is None:
        maxvol = tsize**3 / 5

    dt = 2 * np.pi / ndiv
    theta = np.arange(dt, 2 * np.pi + dt, dt)
    cx = np.outer(np.array(r), np.cos(theta))
    cy = np.outer(np.array(r), np.sin(theta))

    p0 = np.column_stack((cx[0, :], cy[0, :], np.zeros(ndiv)))
    p1 = np.column_stack((cx[1, :], cy[1, :], len_axis * np.ones(ndiv)))

    pp = np.vstack((p0, p1))
    no = rotatevec3d(pp, v0.T) + np.tile(c0.T, (pp.shape[0], 1))

    # face = np.empty((0,4))
    face = []
    for i in range(1, ndiv):
        # face = np.vstack((face, np.array([i, i + ndiv, i + ndiv + 1, i + 1])))
        face.append([[[i, i + ndiv, i + ndiv + 1, i + 1]], [1]])

    face.append([[[ndiv, 2 * ndiv, ndiv + 1, 1]], [1]])
    face.append([[list(range(1, ndiv + 1))], [2]])
    face.append([[list(range(ndiv + 1, 2 * ndiv + 1))], [3]])

    if tsize == 0.0 and maxvol == 0.0:
        return no, face

    node, elem, *_ = surf2mesh(
        no,
        face,
        np.min(no, axis=0),
        np.max(no, axis=0),
        1,
        maxvol,
        regions=np.array([[0, 0, 1]]),
        holes=np.array([]),
    )
    face, *_ = volface(elem[:, 0:4])

    return node, face, elem


# _________________________________________________________________________________________________________


def meshgrid5(*args):
    args = list(args)

    n = len(args)
    if n != 3:
        raise ValueError("only works for 3D case!")

    for i in range(n):
        v = args[i]
        if len(v) % 2 == 0:
            args[i] = np.linspace(v[0], v[-1], len(v) + 1)

    # create a single n-d hypercube
    cube8 = np.array(
        [
            [1, 4, 5, 13],
            [1, 2, 5, 11],
            [1, 10, 11, 13],
            [11, 13, 14, 5],
            [11, 13, 1, 5],
            [2, 3, 5, 11],
            [3, 5, 6, 15],
            [15, 11, 12, 3],
            [15, 11, 14, 5],
            [11, 15, 3, 5],
            [4, 5, 7, 13],
            [5, 7, 8, 17],
            [16, 17, 13, 7],
            [13, 17, 14, 5],
            [5, 7, 17, 13],
            [5, 6, 9, 15],
            [5, 8, 9, 17],
            [17, 18, 15, 9],
            [17, 15, 14, 5],
            [17, 15, 5, 9],
            [10, 13, 11, 19],
            [13, 11, 14, 23],
            [22, 19, 23, 13],
            [19, 23, 20, 11],
            [13, 11, 19, 23],
            [11, 12, 15, 21],
            [11, 15, 14, 23],
            [23, 21, 20, 11],
            [23, 24, 21, 15],
            [23, 21, 11, 15],
            [16, 13, 17, 25],
            [13, 17, 14, 23],
            [25, 26, 23, 17],
            [25, 22, 23, 13],
            [13, 17, 25, 23],
            [17, 18, 15, 27],
            [17, 15, 14, 23],
            [26, 27, 23, 17],
            [27, 23, 24, 15],
            [23, 27, 17, 15],
        ]
    ).T

    # build the complete lattice
    nodecount = [len(arg) for arg in args]

    if any(count < 2 for count in nodecount):
        raise ValueError("Each dimension must be of size 2 or more.")

    node = lattice(*args)

    ix, iy, iz = np.meshgrid(
        np.arange(1, nodecount[0] - 1, 2),
        np.arange(1, nodecount[1] - 1, 2),
        np.arange(1, nodecount[2] - 1, 2),
        indexing="ij",
    )
    ind = np.ravel_multi_index(
        (ix.flatten() - 1, iy.flatten() - 1, iz.flatten() - 1), nodecount, order="F"
    )

    nodeshift = np.array(
        [
            0,
            1,
            2,
            nodecount[0],
            nodecount[0] + 1,
            nodecount[0] + 2,
            2 * nodecount[0],
            2 * nodecount[0] + 1,
            2 * nodecount[0] + 2,
        ]
    )
    nodeshift = np.concatenate(
        (
            nodeshift,
            nodeshift + nodecount[0] * nodecount[1],
            nodeshift + 2 * nodecount[0] * nodecount[1],
        )
    )

    nc = len(ind)
    elem = np.zeros((nc * 40, 4), dtype=int)
    for i in range(nc):
        elem[np.arange(0, 40) + (i * 40), :] = (
            np.reshape(nodeshift[cube8.flatten() - 1], (4, 40)).T + ind[i]
        )

    elem = elem + 1
    elem = meshreorient(node[:, :3], elem[:, :4])[0]

    return node, elem


# _________________________________________________________________________________________________________


def meshgrid6(*args):
    """
    Generate a tetrahedral mesh from an N-D rectangular lattice by splitting
    each hypercube into 6 tetrahedra.

    Parameters:
        v1, v2, v3, ... : array-like
            Numeric vectors defining the lattice in each dimension.
            Each vector must be of length >= 1.

    Returns:
        node : ndarray
            Coordinates of the nodes in the factorial lattice created from (v1, v2, v3, ...).
            Each row corresponds to a node.
        elem : ndarray
            Integer array defining the simplices (tetrahedra) as indices into rows of `node`.

    Notes:
        This function is part of the iso2mesh toolbox (http://iso2mesh.sf.net)
        Originally authored by John D'Errico, with modifications by Qianqian Fang.
    """
    # dimension of the lattice
    n = len(args)

    # create a single n-d hypercube     # list of node of the cube itself
    vhc = (
        np.array(list(map(lambda x: list(bin(x)[2:].zfill(n)), range(2**n)))) == "1"
    ).astype(int)

    # permutations of the integers 1:n
    p = list(permutations(range(1, n + 1)))
    p = p[::-1]
    nt = len(p)
    thc = np.zeros((nt, n + 1), dtype=int)

    for i in range(nt):
        thc[i, :] = np.where(
            np.all(np.diff(vhc[:, np.array(p[i]) - 1], axis=1) >= 0, axis=1)
        )[0]

    # build the complete lattice
    nodecount = np.array([len(arg) for arg in args])
    if np.any(nodecount < 2):
        raise ValueError("Each dimension must be of size 2 or more.")
    node = lattice(*args)

    # unrolled index into each hyper-rectangle in the lattice
    ind = [np.arange(nodecount[i] - 1) for i in range(n)]
    ind = np.meshgrid(*ind, indexing="ij")
    ind = np.array(ind).reshape(n, -1).T
    k = np.cumprod([1] + nodecount[:-1].tolist())

    ind = 1 + ind @ k.T  # k[:-1].reshape(-1, 1)
    nind = len(ind)
    offset = vhc @ k.T
    elem = np.zeros((nt * nind, n + 1), dtype=int)
    L = np.arange(1, nind + 1).reshape(-1, 1)

    for i in range(nt):
        elem[L.flatten() - 1, :] = np.tile(ind, (n + 1, 1)).T + np.tile(
            offset[thc[i, :]], (nind, 1)
        )
        L += nind

    elem = meshreorient(node[:, :3], elem[:, :4])[0]

    return node, elem


# _________________________________________________________________________________________________________


def lattice(*args):
    n = len(args)
    sizes = [len(arg) for arg in args]
    grids = np.meshgrid(*args, indexing="ij")
    grid = np.zeros((np.prod(sizes), n))
    for i in range(n):
        grid[:, i] = grids[i].ravel(order="F")
    return grid


# _________________________________________________________________________________________________________


def latticegrid(*args):
    """
    node, face, centroids = latticegrid(xrange, yrange, zrange, ...)

    Generate a 3D lattice.

    Parameters:
        *args: 1D arrays specifying the range of each dimension.

    Returns:
        node: (N, D) array of node coordinates.
        face: list of faces (each a list of indices starting from 1).
        centroids: (M, D) array of centroid coordinates of each lattice cell.
    """
    n = len(args)
    p = np.meshgrid(*args, indexing="ij")
    node = np.zeros((p[0].size, n))
    for i in range(n):
        node[:, i] = p[i].ravel(order="F")

    if n == 1:
        return node

    dim = p[0].shape
    dd = [dim[0], dim[0] * dim[1]]

    onecube = np.array(
        [
            [0, dd[0], dd[0] + 1, 1],
            [0, 1, dd[1] + 1, dd[1]],
            [0, dd[1], dd[1] + dd[0], dd[0]],
        ]
    )
    onecube = np.vstack(
        [
            onecube,
            onecube + np.array([[dd[1]], [dd[0]], [1]]) @ np.ones((1, 4), dtype=int),
        ]
    )

    len_cube = np.prod(np.array(dim[:3]) - 1)
    face = np.tile(onecube, (len_cube, 1))

    xx, yy, zz = np.meshgrid(
        np.arange(1, dim[0]), np.arange(1, dim[1]), np.arange(1, dim[2]), indexing="ij"
    )

    # Convert subscript to linear index in column-major order (MATLAB-style)
    idx = (
        np.ravel_multi_index(
            (xx.ravel(order="F") - 1, yy.ravel(order="F") - 1, zz.ravel(order="F") - 1),
            dim,
            order="F",
        )
        + 1
    )  # 1-based index for face construction
    orig = np.tile(idx, (onecube.shape[0], 1))

    for i in range(onecube.shape[1]):
        face[:, i] = face[:, i] + orig.ravel(order="F")

    # Convert to 1-based row-unique face list (like MATLAB)
    face = np.unique(face, axis=0)
    face = np.array([list(row) for row in face])

    centroids = None
    if len(args) >= 3:
        diffvec = [np.diff(arg) for arg in args]
        xx, yy, zz = np.meshgrid(*diffvec, indexing="ij")
        centroids = (
            node[idx - 1, :]
            + 0.5
            * np.vstack(
                [xx.ravel(order="F"), yy.ravel(order="F"), zz.ravel(order="F")]
            ).T
        )

    return node, face.tolist(), centroids


# _________________________________________________________________________________________________________


def extrudecurve(
    xy, yz, Nx=30, Nz=30, Nextrap=0, spacing=1, anchor=None, dotopbottom=0
):
    """
    Create a triangular surface mesh by swinging a 2D spline along another 2D spline curve.

    Parameters:
        xy : ndarray
            A 2D spline path, along which the surface is extruded, defined on the x-y plane.
        yz : ndarray
            A 2D spline which will move along the path to form a surface, defined on the y-z plane.
        Nx : int, optional
            The count of sample points along the extrusion path (xy), default is 30.
        Nz : int, optional
            The count of sample points along the curve to be extruded (yz), default is 30.
        Nextrap : int, optional
            Number of points to extrapolate outside of the xy/yz curves, default is 0.
        spacing : float, optional
            Define a spacing scaling factor for spline interpolations, default is 1.
        anchor : list or ndarray, optional
            The 3D point in the extruded curve plane (yz) that is aligned at the nodes long the extrusion path.
            If not provided, it is set as the point on the interpolated yz with the largest y-value.
        dotopbottom : int, optional
            If set to 1, tessellated top and bottom faces will be added, default is 0.

    Returns:
        node : ndarray
            3D node coordinates for the generated surface mesh.
        face : ndarray
            Triangular face patches of the generated surface mesh, each row represents a triangle.
        yz0 : ndarray
            Sliced yz curve at the start.
        yz1 : ndarray
            Sliced yz curve at the end.

    -- this function is part of iso2mesh toolbox (http://iso2mesh.sf.net)
    """
    from scipy.interpolate import splev, splrep

    # Compute interpolation points along the xy curve
    xrange = np.max(xy[:, 0]) - np.min(xy[:, 0])
    dx = xrange / Nx
    xi = np.arange(
        np.min(xy[:, 0]) - Nextrap * dx,
        np.max(xy[:, 0]) + Nextrap * dx + spacing * dx / 2,
        spacing * dx,
    )
    pxy = splrep(xy[:, 0], xy[:, 1])

    # Evaluate the interpolated y values and gradients
    yi = splev(xi, pxy)
    dy = np.gradient(yi)
    dxi = np.gradient(xi)

    nn = np.sqrt(dxi**2 + dy**2)
    normaldir = np.vstack((dxi / nn, dy / nn)).T

    # Compute interpolation points along the yz curve
    zrange = np.max(yz[:, 1]) - np.min(yz[:, 1])
    dz = zrange / Nz
    zi = np.arange(
        np.min(yz[:, 1]) - Nextrap * dz,
        np.max(yz[:, 1]) + Nextrap * dz + spacing * dz / 2,
        spacing * dz,
    )
    pyz = splrep(yz[:, 1], yz[:, 0])

    yyi = splev(zi, pyz)

    # Determine anchor point if not provided
    if anchor is None:
        loc = np.argmax(yyi)
        anchor = [0, yyi[loc], zi[loc]]

    # Initialize node and face arrays
    node = np.zeros((len(zi) * len(xi), 3))
    face = np.zeros((2 * (len(zi) - 1) * (len(xi) - 1), 3), dtype=int)

    # Generate the base yz profile points
    xyz = np.column_stack((np.zeros_like(yyi), yyi, zi))
    for i in range(len(xi)):
        # Compute local rotation matrix
        rot2d = np.array(
            [[normaldir[i, 0], -normaldir[i, 1]], [normaldir[i, 1], normaldir[i, 0]]]
        )
        offset = [xi[i], yi[i], anchor[2]]
        newyz = xyz.copy()
        newyz[:, :2] = (rot2d @ (newyz[:, :2] - anchor[:2]).T).T + offset[:2]
        node[i * len(zi) : (i + 1) * len(zi), :] = newyz

        # Create faces between segments
        if i > 0:
            a = np.arange(len(zi) - 1)
            b = a + 1
            f1 = np.stack(
                (a + (i - 1) * len(zi), a + i * len(zi), b + (i - 1) * len(zi)), axis=-1
            )
            f2 = np.stack(
                (b + (i - 1) * len(zi), a + i * len(zi), b + i * len(zi)), axis=-1
            )
            face[(i - 1) * 2 * (len(zi) - 1) : (i) * 2 * (len(zi) - 1)] = np.vstack(
                (f1, f2)
            )

        # Save yz slices for later output
        if i == Nextrap:
            yz0 = newyz[Nextrap : len(zi) - Nextrap, :]
        if i == len(xi) - Nextrap - 1:
            yz1 = newyz[Nextrap : len(zi) - Nextrap, :]

    # Add two flat polygons on the top and bottom of the contours
    # to ensure the enclosed surface is not truncated by meshfix
    if dotopbottom == 1:
        from scipy.spatial import Delaunay

        C = np.vstack((np.arange(0, len(xi) - 1), np.arange(1, len(xi)))).T
        C = np.vstack((C, [[len(xi) - 1, 0]]))
        dt = Delaunay(np.column_stack((xi, yi)))
        io = dt.find_simplex(np.column_stack((xi, yi))) >= 0
        endface = dt.simplices[io]
        endface = (endface - 1) * len(zi) + 1
        face = np.vstack((face, endface, endface + len(zi) - 1))

    # Check and repair mesh geometry
    node, face = meshcheckrepair(node, face, "deep")

    return node, face, yz0, yz1


# _________________________________________________________________________________________________________


def meshcylinders(c0, v, seglen, r, tsize=None, maxvol=None, ndiv=20):
    """
    create the surface and (optionally) tetrahedral mesh of multiple segments of 3D cylinders

    author: Qianqian Fang, <q.fang at neu.edu>

    Parameters:
        c0: cylinder list axis's starting point
        v: directional vector of the cylinder
        seglen: a scalar or a vector denoting the length of each
             cylinder segment along the direction of v
        args: tsize, maxvol, ndiv - see meshacylinder for details

    Returns:
        node, face, elem - see meshacylinder for details

    -- this function is part of iso2mesh toolbox (http://iso2mesh.sf.net)
    """
    seglen = np.cumsum(seglen)
    c0 = np.array(c0)
    v = np.array(v)
    ncyl, fcyl = meshacylinder(c0, c0 + v * seglen[0], r, 0, 0, ndiv)

    if len(seglen) == 1:
        node = ncyl
        face = fcyl
        return node, face

    for i in range(1, len(seglen)):
        ncyl1, fcyl1 = meshacylinder(
            c0 + v * seglen[i - 1], c0 + v * seglen[i], r, 0, 0, ndiv
        )
        fcyl1 = [[(np.array(f[0]) + ncyl.shape[0]).tolist(), f[1]] for f in fcyl1]
        fcyl1 = fcyl1[:-2] + [fcyl1[-1]]
        fcyl.extend(fcyl1)
        ncyl = np.vstack((ncyl, ncyl1))

    ncyl, I, J = np.unique(
        np.round(ncyl, 10), axis=0, return_index=True, return_inverse=True
    )

    fcyl = [[(J[np.array(f[0]) - 1] + 1).tolist(), f[1]] for f in fcyl]

    if tsize == 0 and maxvol == 0:
        return ncyl, fcyl

    if not tsize:
        tsize = seglen[-1] * 0.1
    if not maxvol:
        maxvol = tsize * tsize * tsize

    centroid = np.cumsum(np.concatenate(([0], seglen[:-1]))) + seglen[-1] * 0.5
    seeds = c0 + v * centroid[:, None]

    node, elem, face = surf2mesh(ncyl, fcyl, None, None, 1, maxvol, seeds, None, 0)
    return node, face, elem


# _________________________________________________________________________________________________________


def extrudesurf(no, fc, vec):
    """
    Create an enclosed surface mesh by extruding an open surface.

    Parameters:
    no : ndarray
        2D array containing the 3D node coordinates of the original surface.
    fc : ndarray
        2D array representing the triangular faces of the original surface.
        Each row corresponds to a triangle defined by indices of 3 nodes.
    vec : array or scalar
        If an array, defines the extrusion direction. If scalar, the normal vector
        is used and multiplied by this scalar for extrusion.

    Returns:
    node : ndarray
        3D node coordinates for the generated surface mesh.
    face : ndarray
        Triangular face patches of the generated surface mesh.
    """

    nlen = no.shape[0]  # Number of nodes in the original surface

    if len(vec) > 1:  # Extrude using a specified vector
        node = np.vstack([no, no + np.tile(vec, (nlen, 1))])
    else:  # Extrude along the surface normal
        node = np.vstack([no, no + vec * nodesurfnorm(no, fc)])

    face = np.vstack([fc, fc + nlen])  # Create top and bottom faces

    # Find surface edges and create side faces
    edge = surfedge(fc)
    sideface = np.hstack([edge, edge[:, [0]] + nlen])
    sideface = np.vstack([sideface, edge + nlen, edge[:, [1]]])

    face = np.vstack([face, sideface])  # Combine all faces

    # Perform mesh repair (fix degenerate elements, etc.)
    node, face = meshcheckrepair(node, face)

    return node, face


# _________________________________________________________________________________________________________


def meshanellip(c0, rr, tsize, maxvol=None):
    """
    Create the surface and tetrahedral mesh of an ellipsoid.

    Parameters:
    c0 : list or ndarray
        Center coordinates [x0, y0, z0] of the ellipsoid.
    rr : list or ndarray
        Radii of the ellipsoid. If rr is:
            - Scalar: a sphere with radius rr.
            - 1x3 or 3x1 vector: specifies the ellipsoid radii [a, b, c].
            - 1x5 or 5x1 vector: specifies [a, b, c, theta, phi], where theta and phi are rotation angles along the z and x axes.
    tsize : float
        Maximum surface triangle size on the ellipsoid.
    maxvol : float, optional
        Maximum volume of the tetrahedral elements.

    Returns:
    node : ndarray
        Node coordinates, 3 columns for x, y, and z respectively.
    face : ndarray
        Surface mesh face elements (each row has 3 vertices).
    elem : ndarray, optional
        Tetrahedral mesh elements (each row has 4 vertices).
    """

    rr = np.asarray(rr).flatten()

    if len(rr) == 1:
        rr = [rr[0], rr[0], rr[0]]  # Sphere case
    elif len(rr) == 3:
        pass  # Already in ellipsoid format
    elif len(rr) != 5:
        raise ValueError("Invalid rr length. See help for details.")

    rmax = min(rr[:3])

    if maxvol is None:
        maxvol = tsize**3  # Set maxvol based on tsize if not provided

    # Call meshunitsphere to generate unit sphere mesh
    if maxvol:
        node, face, elem = meshunitsphere(tsize / rmax, maxvol=maxvol / (rmax**3))
    else:
        node, face = meshunitsphere(tsize / rmax)

    # Scale the unit sphere to the ellipsoid
    node = node @ np.diag(rr[:3])

    if len(rr) == 5:
        theta = rr[3]
        phi = rr[4]

        # Rotation matrices for theta (z-axis) and phi (x-axis)
        Rz = np.array(
            [
                [np.cos(theta), np.sin(theta), 0],
                [-np.sin(theta), np.cos(theta), 0],
                [0, 0, 1],
            ]
        )

        Rx = np.array(
            [[1, 0, 0], [0, np.cos(phi), np.sin(phi)], [0, -np.sin(phi), np.cos(phi)]]
        )

        # Apply rotation to the node coordinates
        node = (Rz @ (Rx @ node.T)).T

    # Translate the ellipsoid to the center c0
    node += np.array(c0).reshape(1, 3)

    return node, face, elem if maxvol else (node, face)



"""@package docstring
Iso2Mesh for Python - Surface mesh processing

Copyright (c) 2024-2025 Qianqian Fang <q.fang at neu.edu>
"""

__all__ = ["sms"]

##====================================================================================
## dependent libraries
##====================================================================================

import numpy as np

##====================================================================================
## implementations
##====================================================================================


def sms(node, face, iter=10, alpha=0.5, method="laplacianhc"):
    """
    Simplified version of surface mesh smoothing.

    Parameters:
    node   : array-like, node coordinates of a surface mesh
    face   : array-like, face element list of the surface mesh
    iter   : int, smoothing iteration number (default: 10)
    alpha  : float, smoothing parameter (default: 0.5)
    method : string, smoothing method (default: 'laplacianhc')

    Returns:
    newnode : array-like, smoothed node coordinates
    """
    conn = meshconn(face, node.shape[0])
    newnode = smoothsurf(node[:, :3], None, conn, iter, alpha, method, alpha)

    return newnode



from setuptools import setup

with open("README.md", "r") as fh:
    readme = fh.read()

setup(
    name="iso2mesh",
    packages=["iso2mesh"],
    version="0.4.2",
    license="GPLv3+",
    description="One-liner 3D Surface and Tetrahedral Mesh Generation Toolbox",
    long_description=readme,
    long_description_content_type="text/markdown",
    author="Qianqian Fang",
    author_email="fangqq@gmail.com",
    maintainer="Qianqian Fang",
    url="https://github.com/NeuroJSON/pyiso2mesh",
    keywords=[
        "Iso2Mesh",
        "Brain2Mesh",
        "JSONLab",
        "EasyH5",
        "Mesh generation",
        "Finite element method",
        "FEM",
        "FEA",
        "MATLAB",
        "Mesh-based Monte Carlo",
        "CGAL",
        "Tetgen",
        "CSG",
        "Boolean",
        "Surface mesh",
        "Tetrahedral mesh",
        "Modeling",
        "Multiphysics",
        "Image processing",
        "NeuroJSON",
        "Binary JSON",
        "JSON",
        "Data format",
    ],
    platforms="any",
    install_requires=["numpy>=1.8.0", "matplotlib", "jdata>=0.7.1", "bjdata"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)
